subroutine ceq_linprog(nx,nb,f,A,b,xm,iret)
!--------------------
  implicit none
  integer, intent(in) :: nx, nb
  real(kind(1.d0)), intent(in) :: f(nx), A(nb,nx), b(nb)
  integer, intent(out) :: iret
  real(kind(1.d0)), intent(out) :: xm(nx)
!--------------------
!  Determine x=xm which minimizes g=f'*x subject to
!  x(i)>=0 and A*x=b, where A has full rank.

! Input:
!	nx	- number of components of x
!	nb	- number of components of b
!	f	- nx-vector f
!	A	- nx x nb matrix A
! Output:
!	xm	- solution	
!	iret= 0 if solution is found 
!	iret=-1 if g is unbounded
!	iret=-2 if there is no feasible solution
!	iret=-3 if A is rank deficient

!  S.B. Pope 10/1/02

integer :: icase, iposv(nb), izrov(nx), j, jj, &
   l1(nb+nx+2),l2(nb+nx+2), nl1, nl2, kp
real(kind(1.d0)) :: T(nb+2,nx+1), eps=1.d-9, q1

T=0.d0
T(1,2:nx+1)=-f
T(2:nb+1,1)=b
T(2:nb+1,2:nx+1)=-A

do j=2,nb+1  !  change sign of row if necessary
	if( T(j,1) < 0.d0 ) T(j,:)=-T(j,:)
end do

call simplex(icase)

if( icase /= 0 ) then
   if( icase > 0 ) iret=-1 ! g is unbounded
   if( icase < 0 ) iret=-2 ! no feasible solution
   return
endif

! unravel solution
	
xm = 0.d0

do j = 1, nb
  if( iposv(j) <= nx+1 ) then	! basic variable determined
    jj = iposv(j)
    xm(jj) = T(j+1,1) ! feasible solution
  else	! equation is linearly dependent
    iret=-3
	return
  endif
end do

iret=0

contains
!----------------------------------------------
!  The following is based on the Numerical Recipes
!	routine simplx.

  subroutine simplex(icase)

  implicit none
  integer, intent(out) :: icase

  integer :: k, i, l3(nb+nx+2),kh, ir, ip, is
  real(kind(1.d0)) :: bmax


  nl1=nx
  do k=1,nx
    l1(k)=k
    izrov(k)=k
  end do

  nl2=nb
  do i=1,nb
    l2(i)=i
    iposv(i)=nx+i
  end do
  
  ir=0
  if(nb == 0) go to 30

  ir=1
  T(nb+2,1:nx+1)=-sum(T(2:nb+1,1:nx+1),1)
  
10  continue
  call sub_a(nb+1,0,kp,bmax)

  if(bmax <= eps .and. T(nb+2,1) < -eps)then
    icase=-1
    return
  else if(bmax <= eps .and. T(nb+2,1) <= eps)then
    
    if(1 <= nb)then
      do ip=1,nb
        if(iposv(ip) == ip+nx)then
          call sub_a(ip,1,kp,bmax)
          if(bmax > 0.d0)go to 1
        endif
      end do
    endif
    ir=0
    
    go to 30
  endif

  call sub_b(ip,q1)

  if(ip == 0)then
    icase=-1
    return
  endif

1  continue
  call sub_c(nb+1,nx,ip+1,kp+1)

  if(iposv(ip) >= nx+1)then
    do k=1,nl1
      if(l1(k) == kp)go to 2
    end do

2   nl1=nl1-1

    do is=k,nl1
      l1(is)=l1(is+1)
    end do
  else
    if(iposv(ip) < nx+1) go to 20
    kh=iposv(ip)-nx
    if(l3(kh) == 0) go to 20
    l3(kh)=0
  endif

  T(nb+2,kp+1)=T(nb+2,kp+1)+1.
  T(1:nb+2,kp+1)=-T(1:nb+2,kp+1)

20  continue
  is=izrov(kp)
  izrov(kp)=iposv(ip)
  iposv(ip)=is
  if(ir /= 0) go to 10

30  continue
  call sub_a(0,0,kp,bmax)

  if(bmax <= 0.d0)then
    icase=0  !  solution found
    return
  endif

  call sub_b(ip,q1)
  if(ip == 0)then
    icase=1
    return
  endif

  call sub_c(nb,nx,ip+1,kp+1)
  go to 20
  
end subroutine simplex

!------------------------------------------------------------
  subroutine sub_a(mm,iabf,kp,bmax)

  implicit none
  integer, intent(in) :: mm, iabf
  integer, intent(out) :: kp
  real(kind(1.d0)), intent(out) :: bmax

  integer :: k
  real(kind(1.d0)) :: test
  
  kp=l1(1)
  bmax=T(mm+1,kp+1)
 
  do k=2,nl1
    if(iabf == 0)then
      test=T(mm+1,l1(k)+1)-bmax
    else
      test=abs(T(mm+1,l1(k)+1))-abs(bmax)
    endif
    if(test > 0.d0) then
      bmax=T(mm+1,l1(k)+1)
      kp=l1(k)
    endif
  end do

  return
  end subroutine sub_a
!------------------------------------------------------------
  subroutine sub_b(ip,q1)

  implicit none
  integer, intent(out) :: ip
  real(kind(1.d0)), intent(out) :: q1

  integer :: kp1, i, k, ii
  real(kind(1.d0)) :: q, q0, qp
  
  ip=0
  kp1=kp+1

  do i=1,nl2
    if(T(l2(i)+1,kp1) < -eps) go to 2
  end do
  return

2 q1=-T(l2(i)+1,1)/T(l2(i)+1,kp1)
  ip=l2(i)

  do i=i+1,nl2
    ii=l2(i)
    if(T(ii+1,kp1).lt.-eps)then
      q=-T(ii+1,1)/T(ii+1,kp1)
      if(q < q1)then
        ip=ii
        q1=q
      else if (q == q1) then
        do k=1,nx
          qp=-T(ip+1,k+1)/T(ip+1,kp1)
          q0=-T(ii+1,k+1)/T(ii+1,kp1)
          if(q0 /= qp)go to 6
        end do
6       if(q0 < qp) ip=ii
      endif
    endif
  end do

  return
  end subroutine sub_b
!------------------------------------------------------------
  subroutine sub_c(i1,k1,ip1,kp1)

  implicit none
  integer, intent(in) :: i1,k1,ip1,kp1

  integer :: ii, kk
  real(kind(1.d0)) :: piv, Tp

  piv=1.d0/T(ip1,kp1)

  do ii=1,i1+1
    if(ii /= ip1 ) then
      T(ii,kp1)=T(ii,kp1)*piv
	  Tp=T(ii,kp1)
      T(ii,1:k1+1)=T(ii,1:k1+1)-T(ip1,1:k1+1)*Tp
	  T(ii,kp1)=Tp
    endif
  end do

  T(ip1,1:k1+1) = -T(ip1,1:k1+1)*piv
  T(ip1,kp1)=piv

  return
  end subroutine sub_c

end subroutine ceq_linprog