program ceq_test

!  Test the constrained chemical equilibrium algorithm.
!  S. B. Pope  7/12/03


use ceq_system

implicit none
integer, parameter :: ns=49, ne=5, nth=15, nsamp=1100, ndi=2*ns+1, &
                      ncs=5, ng=20, nc=ncs+ne+ng 

integer :: CS(ncs), i, iret, k, info, lwork, ntry, lu, diag,  &
           nn, is, itemps, itimes, inewts, iys, ngm

real(kind(1.d0)) :: p_atm, Ein(ns,ne), thermo(ns,nth), DI(nsamp,ndi), T, &
	 p0_Pa, p, BB(ns,ns), SBB(ns), UBB(ns,ns), &
	 VBBT(ns,ns), work(4*ns*ns), Bg(ns,ng), &
	 stats(2*nsamp,8), Tin, hoRin, zin(ns), Teq, &
	 zeq(ns), stat(20), zguess(ns), cb(nc), HoR,  err

logical :: test_fixed_T, test_fixed_h

real :: cpu_start, cpu_end

type (sys_type), pointer :: sys


!--- specify conditions and constraints

p_atm=1.d0       ! pressure in atmospheres
CS=(/ (i,i=1,ncs) /)      ! list of species to be constrained


!--- read element, thermo and test composition data

call ceq_input(ns,ne,nth,nsamp,ndi,Ein,thermo,DI)

!  form general constraints (orthogonal to other constraints)
BB=0.d0
BB(:,1:ne)=Ein
do k=1, min(ncs,ns-ne)
    BB(CS(k),ne+k)=1.d0
end do

lwork=size(work)
call dgesvd('A','A',ns,ns,BB(1:ns,1:ns),ns,SBB(1:ns), &
   UBB(1:ns,1:ns),ns,VBBT(1:ns,1:ns),ns,work(1:lwork),lwork,info)

if( info /=0 ) then
  write(0,*)'ceq_test: svd failed, info= ', info
  stop
endif

!  set general constraints to the columns of UBB, starting with the independent ones

do i = 1, ng
   k = mod( ne+ncs+i-1 , ns ) + 1
   Bg(:,i) = UBB(:,k)
end do

!  modify constraints to add linear dependence (as a test)
if( ng>=4 ) then
    Bg(:,     ng-1)=0.d0      ! make column dependent on constrained species
    Bg(CS(1), ng-1)=1.d0       
    Bg(:,       ng)=0.d0      ! make last species determined
    Bg(ns,      ng)=1.d0
    Bg(:,     ng-2)=Ein(:,1)  ! make column dependent on elements
	Bg(:,     ng-3)=Bg(:,1)   ! make column dependent on 1st general constraint
endif

!---------------------  initialize system -------------------------------
lu   = 0  ! logical unit number for outpur
diag = 5  ! level of diagnostics

call ceq_sys_init( ns ,ne, ncs, ng, Ein, CS, Bg, thermo, lu, diag, sys, info )

if( info /= 0 ) then
   write(0,*) 'ceq_test: ceq_sys_init failed, info = ', info
   stop
endif

!  change parameter settings

call ceq_param_set( sys, diag=1, eps_sp=1.d-9, srat_lim=1.d-6, dec_min=0.9d0, res_tol=1.d-9 )

!----  specify problem to solve and parameters

test_fixed_T=.true. ! for fixed T calculation
test_fixed_h=.true. ! for fixed h calculation

!---  loop over test compositions

stats= 0.d0
nn   = 0
ntry = 100
call cpu_time( cpu_start )

do is = 1, ntry
    T     = DI(is,1)
    Tin   = T
    hoRin = 0.d0
    zin   = DI(is,2:ns+1)

    if( test_fixed_T ) then
       call ceq_state( sys,  p_atm=p_atm, N=zin, T=Tin,  &
                      T_eq=Teq, N_eq=zeq, HoR_eq = HoR, stats=stat, info=info )

       if( info < 0 ) then
	      write(0,*) 'ceq_test: 1st call to ceq_state failed, info = ', info
          stop
       endif

       nn=nn+1
       stats(nn,1:8)=stat(1:8)

       if( diag >=2 ) then
          itemps = nint( stat(1) )
          itimes = nint( stat(2) )
          inewts = nint( stat(4) )
          iys    = nint( stat(5) )
          err    = stat(6)
          write(0,600) is, itemps, itimes, inewts, iys, Tin, Teq, Tin-Teq, err
	   endif
	endif

    if( test_fixed_h ) then
       call ceq_state( sys,  p_atm=p_atm, N=zin, T_h=Tin, N_h=zin, T_g=Tin,  &
                      T_eq=Teq, N_eq=zeq, HoR_eq = HoR, stats=stat, info=info )

       if( info < 0 ) then
	      write(0,*) 'ceq_test: 2nd call to ceq_state failed, info = ', info
          stop
       endif

       nn=nn+1
       stats(nn,1:8)=stat(1:8)

       if( diag >=2 ) then
          itemps = nint( stat(1) )
          itimes = nint( stat(2) )
          inewts = nint( stat(4) )
          iys    = nint( stat(5) )
          err    = stat(6)
          write(0,600) is, itemps, itimes, inewts, iys, Tin, Teq, Tin-Teq, err
	   endif
	endif
end do

call cpu_time(cpu_end)

write(0,*)' '
write(0,*)'Average counts'
write(0,*)'Temp  its  = ', sum(stats(1:ntry,1))/nn
write(0,*)'Time steps = ', sum(stats(1:ntry,2))/nn
write(0,*)'Newton its = ', sum(stats(1:ntry,4))/nn
write(0,*)'Func evals = ', sum(stats(1:ntry,5))/nn
write(0,'(a,1pe13.4)')'CPU time (ms) = ', (cpu_end-cpu_start)*1.e3/nn

pause
stop

600  format(5i6,1p,104e12.4)
end program ceq_test