program demo1

! demo1: Demonstration program for the CEQ library.
! Comments pertaining to the use of CEQ are indicated by !CEQ.

! The chemical equilibrium composition is calculated for fuel(H2/N2)-oxidant(O2/N2)
! mixtures as a function of mixture fraction, at fixed pressure and enthalpy.

! The species considered are:
! H2 H O2 O OH HO2 H2O2 H2O N2
! 1  2 3  4 5  6   7    8   9

!CEQ: the following "use" statement is required in every routine calling CEQ routines
use ceq_system 
 
implicit none

type (sys_type), pointer, save :: sys  !CEQ: required declaration for the structure sys

integer, parameter :: ne=3, ns=9, nxi=100, lu=8
integer :: iret, iret1, iret2, CS(1), i, ncs, ng, lu_op, diag

real(kind(1.d0)) :: xfu(ns), xox(ns), Tfu, Tox, patm, Ein(ns,ne), thermo(ns,15), Bg(1,1), &
                    Teq, xi, stats(20), xeq(ns), wt(ns), &
					z(ns), zfu(ns), zox(ns), zeq(ns), HoR_fu, HoR_ox, HoR

!--------  specify conditions -----------------------------------

patm   = 1.d0  !  pressure (atm)

xfu    = 0.d0  !  species (relative volume units) and temperature (K) of pure fuel
xfu(1) = 0.5d0 
xfu(9) = 0.5d0 
Tfu    = 300.d0

xox    = 0.d0  !  species (relative volume units) and temperature (K) of pure oxidant
xox(3) = 0.21d0 
xox(9) = 0.79d0 
Tox    = 1000.d0

!------------  read thermo data from demo1.dat -----------------

open( lu, file='demo1.dat', form='formatted', err=110 )
read( lu,10) wt       ! molecular weights
read( lu,10) Ein      ! element matrix
read( lu,10) thermo   ! thermo data
close(lu)
10    format((1p,5e25.15))

!CEQ---------  initialize CEQ --------------------------------------

ncs   = 0 ! no constrained species
ng    = 0 ! no general linear constraints
lu_op = 0 ! logical unit for output (used also by ceq_state)
diag  = 5 ! level of diagnostic output

call ceq_sys_init( ns, ne, ncs, ng, Ein, CS, Bg, thermo, lu_op, diag, sys, iret )
if( iret < 0 ) then
   write(0,*)'demo1: error code on return from ceq_sys_init, iret = ', iret
   stop
endif

!CEQ:  subroutine ceq_param_set can be called (after ceq_sys_init) to change parameters

call ceq_param_set( sys, diag=1 )  !  reduce diagnostic level

open( lu, file='demo_out.1', form='formatted', err=100 )

!--------  determine properties of pure fuel and oxidant ---------------

zfu = xfu / sum(xfu*wt)  ! specific moles of species in fuel
zox = xox / sum(xox*wt)  ! specific moles of species in oxidant

!  determine normalized enthalpy of fuel and oxidant.
!  (the use of ceq_state for this purpose is simple but inefficient.)

call ceq_state( sys, N=zfu, p_atm=patm, N_h=zfu, T_h=Tfu, HoR_eq=HoR_fu, info=iret1 )
call ceq_state( sys, N=zox, p_atm=patm, N_h=zox, T_h=Tox, HoR_eq=HoR_ox, info=iret2 )

if( iret1 < 0  .or.  iret2 < 0 ) then  !CEQ negative value indicates failure
   write(0,*)'demo1: error code on return from ceq_state, iret1, iret2 = ', iret1, iret2
   stop
endif

!--------  loop over mixture fractions  -----------------------------------

do i   = 1, nxi
   xi  = (i-1)/float(nxi-1)                ! mixture fraction
   z   = zfu    * xi + zox    * (1.d0-xi)  ! specific moles of species
   HoR = HoR_fu * xi + HoR_ox * (1.d0-xi)  ! normalized enthalpy

!CEQ: determine chemical equilibrium at specified p and h.
!CEQ: zeq contains the specific moles of species at equilibrium. 

   call ceq_state( sys, N=z, p_atm=patm, HoR=HoR, &
                   N_eq=zeq, T_eq=Teq, stats=stats, info=iret )

   if( iret < 0 ) then
      write(0,*)'demo1: error code on return from ceq_state, iret = ', iret
	  stop
   endif
   
   xeq = zeq/sum(zeq)  !  form mole fractions
   write(lu,'(1p,15e20.10)') xi, Teq, stats(5), xeq
enddo

write(0,*) 'demo1: normal termination'
stop  !  normal termination

100   write(0,*) 'demo1:  error opening demo_out.1'
stop
110   write(0,*) 'demo1:  error opening demo1.dat'
stop

end program demo1