program demo2

! demo2: Demonstration program for the CEQ library.
! Demonstration of the use of CEQ in conjunction with Chemkin II.
! Comments pertaining to the use of CEQ are indicated by !CEQ.

! The chemical equilibrium composition is calculated for fuel(H2/N2)-oxidant(O2/N2)
! mixtures as a function of mixture fraction, at fixed pressure and enthalpy.

! The species considered are:
! H2 H O2 O OH HO2 H2O2 H2O N2
! 1  2 3  4 5  6   7    8   9

!CEQ: the following "use" statement is required in every routine calling CEQ routines
use ceq_system 
 
implicit none

type (sys_type), pointer, save :: sys  !CEQ: required declaration for the structure sys

integer, parameter :: ne=3, ns=9, nxi=100, lu=8
integer :: nex, nsx, nr, nfit, iret, iret1, iret2, CS(1), i, ncs, ng, lu_op, diag

real(kind(1.d0)), allocatable :: rwk(:)
integer,          allocatable :: iwk(:)
character(16),    allocatable :: cwk(:)

real(kind(1.d0)) :: xfu(ns), xox(ns), Tfu, Tox, patm, Ein(ns,ne), thermo(ns,15), Bg(1,1), &
                    Teq, xi, stats(20), xeq(ns), wt(ns), &
					z(ns), zfu(ns), zox(ns), zeq(ns), HoR_fu, HoR_ox, HoR

!--------  specify conditions -----------------------------------

patm   = 1.d0  !  pressure (atm)

xfu    = 0.d0  !  species (relative volume units) and temperature (K) of pure fuel
xfu(1) = 0.5d0 
xfu(9) = 0.5d0 
Tfu    = 300.d0

xox    = 0.d0  !  species (relative volume units) and temperature (K) of pure oxidant
xox(3) = 0.21d0 
xox(9) = 0.79d0 
Tox    = 1000.d0

!--------  initialize Chemkin and get thermo data ---------------

call ck_init  ! initialize Chemkin and its work arrays

call ckindx( iwk, rwk, nex, nsx, nr, nfit )  ! check indexes
if( nex /= ne .or. nsx /= ns .or. nfit /= 7 ) then
   write(0,*)'demo2: index mismatch, nex, nsx, nfit = ', nex, nsx, nfit
   stop
endif

!CEQ:  subroutine ceq_ck obtains the arrays Ein and thermo from Chemkin II
call ceq_ck( ns, ne, Ein, thermo, iwk, rwk )  ! obtain Ein and thermo
call ckwt( iwk, rwk, wt )  !  obtain molecular weights

!------------  output thermo data for use by demo1 -----------------

open( lu, file='demo1.dat', form='formatted', err=110 )
write(lu,10) wt
write(lu,10) Ein
write(lu,10) thermo
close(lu)
10    format((1p,5e25.15))

!CEQ---------  initialize CEQ --------------------------------------

ncs   = 0 ! no constrained species
ng    = 0 ! no general linear constraints
lu_op = 0 ! logical unit for output (used also by ceq_state)
diag  = 5 ! level of diagnostic output

call ceq_sys_init( ns, ne, ncs, ng, Ein, CS, Bg, thermo, lu_op, diag, sys, iret )
if( iret < 0 ) then
   write(0,*)'demo2: error code on return from ceq_sys_init, iret = ', iret
   stop
endif

!CEQ:  subroutine ceq_param_set can be called (after ceq_sys_init) to change parameters

call ceq_param_set( sys, diag=1 )  !  reduce diagnostic level

open( lu, file='demo_out.2', form='formatted', err=100 )

!--------  determine properties of pure fuel and oxidant ---------------

zfu = xfu / sum(xfu*wt)  ! specific moles of species in fuel
zox = xox / sum(xox*wt)  ! specific moles of species in oxidant

!  determine normalized enthalpy of fuel and oxidant.
!  (the use of ceq_state for this purpose is simple but inefficient.)

call ceq_state( sys, N=zfu, p_atm=patm, N_h=zfu, T_h=Tfu, HoR_eq=HoR_fu, info=iret1 )
call ceq_state( sys, N=zox, p_atm=patm, N_h=zox, T_h=Tox, HoR_eq=HoR_ox, info=iret2 )

if( iret1 < 0  .or.  iret2 < 0 ) then  !CEQ negative value indicates failure
   write(0,*)'demo2: error code on return from ceq_state, iret1, iret2 = ', iret1, iret2
   stop
endif

!--------  loop over mixture fractions  -----------------------------------

do i   = 1, nxi
   xi  = (i-1)/float(nxi-1)                ! mixture fraction
   z   = zfu    * xi + zox    * (1.d0-xi)  ! specific moles of species
   HoR = HoR_fu * xi + HoR_ox * (1.d0-xi)  ! normalized enthalpy

!CEQ: determine chemical equilibrium at specified p and h.
!CEQ: zeq contains the specific moles of species at equilibrium. 

   call ceq_state( sys, N=z, p_atm=patm, HoR=HoR, &
                   N_eq=zeq, T_eq=Teq, stats=stats, info=iret )

   if( iret < 0 ) then
      write(0,*)'demo2: error code on return from ceq_state, iret = ', iret
	  stop
   endif
   
   xeq = zeq/sum(zeq)  !  form mole fractions
   write(lu,'(1p,15e20.10)') xi, Teq, stats(5), xeq
enddo

write(0,*) 'demo2: normal termination'
stop  !  normal termination

100   write(0,*) 'demo2:  error opening demo_out.2'
stop
110   write(0,*) 'demo2:  error opening demo1.dat'
stop

contains !-------------------------------------------------------------------

subroutine ck_init  !  internal subroutine to initialize Chemkin -----------

implicit none
integer :: luck, iflag, luout, liwk, lrwk, lcwk
logical :: exists

inquire( file = 'chem.bin', exist = exists ) ! check that chem.bin exists
if( .not.exists ) then
   write(0,*) 'ck_init: chem.bin does not exist; create using ckinterp'
   stop
endif

luck  = 7  !  open chem.bin on unit luck=7
luout = 0  !  error output on unit 0
open( luck, file='chem.bin', form='unformatted', err=100, action='read', position='rewind'  )

call cklen( luck, luout, liwk, lrwk, lcwk, iflag )  ! obtain array lengths
if( iflag /= 0 ) then
   write(0,*)'ck_init: error code on return from cklen'
   stop
endif

allocate( iwk(liwk) )  !  allocate arrays for Chemkin data
allocate( rwk(lrwk) )
allocate( cwk(lcwk) ) 

call ckinit( liwk, lrwk, lcwk, luck, luout, iwk, rwk, cwk, iflag )
if( iflag /= 0 ) then
   write(0,*) 'ck_init: error code on return from ckinit'
   stop
endif 

close(luck)
return 

100	   write(0,*)'ck_init: error opening chem.bib'
stop

end subroutine ck_init

end program demo2